
### NOT working
import cv2
import os
import time
import math
import mediapipe as mp
import numpy as np
import matplotlib.pyplot as plt


# Initializing mediapipe pose class.
mp_pose = mp.solutions.pose

# Setting up the Pose function for image.
# pose = mp_pose.Pose(static_image_mode=True, min_detection_confidence=0.3, model_complexity=2)

# Setup Pose function for video.
pose_video = mp_pose.Pose(static_image_mode=False, min_detection_confidence=0.6, model_complexity=1)

# Initializing media pipe drawing class, useful for annotation.
mp_drawing = mp.solutions.drawing_utils

# It show colorful hands landmarks
mp_drawing_styles = mp.solutions.drawing_styles

# Initializing media pipe segmentation class.
mp_selfie_segmentation = mp.solutions.selfie_segmentation

# Setting up Segmentation function.
segment = mp_selfie_segmentation.SelfieSegmentation(0)


class FaceDetector():
    def __init__(self, minDetectionCon=0.5):

        self.minDetectionCon = minDetectionCon
        self.mpFaceDetection = mp.solutions.face_detection
        self.mpDraw = mp.solutions.drawing_utils
        self.faceDetection = self.mpFaceDetection.FaceDetection(self.minDetectionCon)

    def findFaces(self, img, draw=True):
        imgRGB = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
        self.results = self.faceDetection.process(imgRGB)
        # print(self.results)
        bboxs = []
        if self.results.detections:
            for id, detection in enumerate(self.results.detections):
                bboxC = detection.location_data.relative_bounding_box
                ih, iw, ic = img.shape
                bbox = int(bboxC.xmin * iw), int(bboxC.ymin * ih), \
                       int(bboxC.width * iw), int(bboxC.height * ih)
                bboxs.append([id, bbox, detection.score])
                if draw:
                    img = self.fancyDraw(img,bbox)

                    cv2.putText(img, f'Probability={int(detection.score[0] * 100)}%',
                            (bbox[0], bbox[1] - 20), cv2.FONT_HERSHEY_PLAIN,2, (255, 0, 255), 2)
        return img, bboxs

    def fancyDraw(self, img, bbox, l=30, t=5, rt= 1):
        x, y, w, h = bbox
        x1, y1 = x + w, y + h

        cv2.rectangle(img, bbox, (255, 0, 255), rt)
        # Top Left  x,y
        cv2.line(img, (x, y), (x + l, y), (255, 0, 255), t)
        cv2.line(img, (x, y), (x, y+l), (255, 0, 255), t)
        # Top Right  x1,y
        cv2.line(img, (x1, y), (x1 - l, y), (255, 0, 255), t)
        cv2.line(img, (x1, y), (x1, y+l), (255, 0, 255), t)
        # Bottom Left  x,y1
        cv2.line(img, (x, y1), (x + l, y1), (255, 0, 255), t)
        cv2.line(img, (x, y1), (x, y1 - l), (255, 0, 255), t)
        # Bottom Right  x1,y1
        cv2.line(img, (x1, y1), (x1 - l, y1), (255, 0, 255), t)
        cv2.line(img, (x1, y1), (x1, y1 - l), (255, 0, 255), t)
        return img

    def make_folder(self, myFolder):
        if not os.path.exists(myFolder):
            os.makedirs(myFolder)
            # print('Creating subfolder: Tim_folder')
            # print('@============ A new sub-folder created for you, sir! =================@')
        return myFolder

    def modifyBackground(self, image, background_image=255, blur=75, threshold=0.75,
                         display=True, method='changeBackground'):
        '''
        This function will replace, blur, desature or make the background transparent depending upon the passed arguments.
        Args:
            image: The input image with an object whose background is required to modify.
            background_image: The new background image for the object in the input image.
            threshold: A threshold value between 0 and 1 which will be used in creating a binary mask of the input image.
            display: A boolean value that is if true the function displays the original input image and the resultant image
                     and returns nothing.
            method: The method name which is required to modify the background of the input image.
        Returns:
            output_image: The image of the object from the input image with a modified background.
            binary_mask_3: A binary mask of the input image.
        '''

        # Convert the input image from BGR to RGB format.
        RGB_img = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)

        # Perform the segmentation.
        result = segment.process(RGB_img)

        # Get a binary mask having pixel value 1 for the object and 0 for the background.
        # Pixel values greater than the threshold value will become 1 and the remainings will become 0.
        binary_mask = result.segmentation_mask
        # threshold = 0.9

        # Stack the same mask three times to make it a three channel image.
        # binary_mask_3 = np.dstack((binary_mask, binary_mask, binary_mask)) # ## NOT working ???
        binary_mask_3 = np.stack((result.segmentation_mask,) * 3, axis=-1) > threshold

        if method == 'changeBackground':

            # Resize the background image to become equal to the size of the input image.
            background_image = cv2.resize(background_image, (image.shape[1], image.shape[0]))

            # Create an output image with the pixel values from the original sample image at the indexes where the mask have
            # value 1 and replace the other pixel values (where mask have zero) with the new background image.
            output_image = np.where(binary_mask_3, image, background_image)

        elif method == 'blurBackground':

            # Create a blurred copy of the input image.
            blurred_image = cv2.GaussianBlur(image, (blur, blur), 0)

            # Create an output image with the pixel values from the original sample image at the indexes where the mask have
            # value 1 and replace the other pixel values (where mask have zero) with the new background image.
            output_image = np.where(binary_mask_3, image, blurred_image)

        elif method == 'desatureBackground':

            # Create a gray-scale copy of the input image.
            grayscale = cv2.cvtColor(src=image, code=cv2.COLOR_BGR2GRAY)

            # Stack the same grayscale image three times to make it a three channel image.
            grayscale_3 = np.dstack((grayscale, grayscale, grayscale))

            # Create an output image with the pixel values from the original sample image at the indexes where the mask have
            # value 1 and replace the other pixel values (where mask have zero) with the new background image.
            output_image = np.where(binary_mask_3, image, grayscale_3)

        elif method == 'transparentBackground':

            # Stack the input image and the mask image to get a four channel image.
            # Here the mask image will act as an alpha channel.
            # Also multiply the mask with 255 to convert all the 1s into 255.
            # output_image = np.dstack((image, binary_mask * 255)) # ## NOT working
            # Create the output image to have white background where ever black is present in the mask.
            output_image = np.where(binary_mask_3, image, 255)


        else:
            # Display the error message.
            print('Invalid Method')

            # Return
            return

        # Check if the original input image and the resultant image are specified to be displayed.
        if display:

            # Display the original input image and the resultant image.
            plt.figure(figsize=[22, 22])
            plt.subplot(121);
            plt.imshow(image[:, :, ::-1]);
            plt.title("Original Image");
            plt.axis('off');
            plt.subplot(122);
            plt.imshow(output_image[:, :, ::-1]);
            plt.title("Output Image");
            plt.axis('off');
            plt.show()
        # Otherwise
        else:

            # Return the output image and the binary mask.
            # Also convert all the 1s in the mask into 255 and the 0s will remain the same.
            # The mask is returned in case you want to troubleshoot.
            return output_image, (binary_mask_3 * 255).astype('uint8')

    def detectPose(image, pose, draw=True, display=True):
        '''
        This function performs pose detection on an image.
        Args:
            image: The input image with a prominent person whose pose landmarks needs to be detected.
            pose: The pose setup function required to perform the pose detection.
            display: A boolean value that is if set to true the function displays the original input image, the resultant image,
                     and the pose landmarks in 3D plot and returns nothing.
        Returns:
            output_image: The input image with the detected pose landmarks drawn.
            landmarks: A list of detected landmarks converted into their original scale.
        '''

        # Create a copy of the input image.
        output_image = image.copy()

        # Convert the image from BGR into RGB format.
        imageRGB = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)

        # Perform the Pose Detection.
        results = pose.process(imageRGB)

        # Retrieve the height and width of the input image.
        height, width, _ = image.shape

        # Initialize a list to store the detected landmarks.
        landmarks = []

        # Check if any landmarks are detected.
        if results.pose_landmarks and draw:

            # Draw Pose landmarks on the output image.
            mp_drawing.draw_landmarks(image=output_image, landmark_list=results.pose_landmarks,
                                      connections=mp_pose.POSE_CONNECTIONS,
                                      # landmark_drawing_spec=mp_drawing_styles.get_default_pose_landmarks_style(), #  colorful landmarks.
                                      landmark_drawing_spec=mp_drawing.DrawingSpec(color=(255, 0, 0), thickness=5,
                                                                                   circle_radius=8),
                                      connection_drawing_spec=mp_drawing.DrawingSpec(color=(49, 125, 237), thickness=5,
                                                                                     circle_radius=2)
                                      )

            # Iterate over the detected landmarks.
            for landmark in results.pose_landmarks.landmark:
                # Append the landmark into the list.
                landmarks.append((int(landmark.x * width), int(landmark.y * height),
                                  (landmark.z * width)))

        # Check if the original input image and the resultant image are specified to be displayed.
        if display:

            # Display the original input image and the resultant image.
            plt.figure(figsize=[22, 22])
            plt.subplot(121);
            plt.imshow(image[:, :, ::-1]);
            plt.title("Original Image");
            plt.axis('off');
            plt.subplot(122);
            plt.imshow(output_image[:, :, ::-1]);
            plt.title("Output Image");
            plt.axis('off');

            # Also Plot the Pose landmarks in 3D.
            mp_drawing.plot_landmarks(results.pose_world_landmarks, mp_pose.POSE_CONNECTIONS)

        # Otherwise
        else:

            # Return the output image and the found landmarks.
            return output_image, landmarks

    def calculateAngle(landmark1, landmark2, landmark3):
        '''
        This function calculates angle between three different landmarks.
        Args:
            landmark1: The first landmark containing the x,y and z coordinates.
            landmark2: The second landmark containing the x,y and z coordinates.
            landmark3: The third landmark containing the x,y and z coordinates.
        Returns:
            angle: The calculated angle between the three landmarks.

        '''

        # Get the required landmarks coordinates.
        x1, y1, _ = landmark1
        x2, y2, _ = landmark2
        x3, y3, _ = landmark3

        # Calculate the angle between the three points
        angle = math.degrees(math.atan2(y3 - y2, x3 - x2) - math.atan2(y1 - y2, x1 - x2))

        # Check if the angle is less than zero.
        if angle < 0:
            # Add 360 to the found angle.
            angle += 360

        # Return the calculated angle.
        return angle

    def classifyPose(landmarks, output_image, display=False):
        '''
        This function classifies yoga poses depending upon the angles of various body joints.
        Args:
            landmarks: A list of detected landmarks of the person whose pose needs to be classified.
            output_image: A image of the person with the detected pose landmarks drawn.
            display: A boolean value that is if set to true the function displays the resultant image with the pose label
            written on it and returns nothing.
        Returns:
            output_image: The image with the detected pose landmarks drawn and pose label written.
            label: The classified pose label of the person in the output_image.

        '''

        # Initialize the label of the pose. It is not known at this stage.
        label = 'Unknown Pose'
        font_2 = cv2.FONT_HERSHEY_COMPLEX
        # Specify the color (Red) with which the label will be written on the image.
        color = (0, 0, 255)

        # Calculate the required angles.
        # ----------------------------------------------------------------------------------------------------------------

        # Get the angle between the left shoulder, elbow and wrist points.
        left_elbow_angle = calculateAngle(landmarks[mp_pose.PoseLandmark.LEFT_SHOULDER.value],
                                          landmarks[mp_pose.PoseLandmark.LEFT_ELBOW.value],
                                          landmarks[mp_pose.PoseLandmark.LEFT_WRIST.value])

        # Get the angle between the right shoulder, elbow and wrist points.
        right_elbow_angle = calculateAngle(landmarks[mp_pose.PoseLandmark.RIGHT_SHOULDER.value],
                                           landmarks[mp_pose.PoseLandmark.RIGHT_ELBOW.value],
                                           landmarks[mp_pose.PoseLandmark.RIGHT_WRIST.value])

        # Get the angle between the left elbow, shoulder and hip points.
        left_shoulder_angle = calculateAngle(landmarks[mp_pose.PoseLandmark.LEFT_ELBOW.value],
                                             landmarks[mp_pose.PoseLandmark.LEFT_SHOULDER.value],
                                             landmarks[mp_pose.PoseLandmark.LEFT_HIP.value])

        # Get the angle between the right hip, shoulder and elbow points.
        right_shoulder_angle = calculateAngle(landmarks[mp_pose.PoseLandmark.RIGHT_HIP.value],
                                              landmarks[mp_pose.PoseLandmark.RIGHT_SHOULDER.value],
                                              landmarks[mp_pose.PoseLandmark.RIGHT_ELBOW.value])

        # Get the angle between the left hip, knee and ankle points.
        left_knee_angle = calculateAngle(landmarks[mp_pose.PoseLandmark.LEFT_HIP.value],
                                         landmarks[mp_pose.PoseLandmark.LEFT_KNEE.value],
                                         landmarks[mp_pose.PoseLandmark.LEFT_ANKLE.value])

        # Get the angle between the right hip, knee and ankle points
        right_knee_angle = calculateAngle(landmarks[mp_pose.PoseLandmark.RIGHT_HIP.value],
                                          landmarks[mp_pose.PoseLandmark.RIGHT_KNEE.value],
                                          landmarks[mp_pose.PoseLandmark.RIGHT_ANKLE.value])

        # ----------------------------------------------------------------------------------------------------------------

        # Check if it is the warrior II pose or the T pose.
        # As for both of them, both arms should be straight and shoulders should be at the specific angle.
        # ----------------------------------------------------------------------------------------------------------------

        # Check if the both arms are straight.
        if left_elbow_angle > 165 and left_elbow_angle < 195 and right_elbow_angle > 165 and right_elbow_angle < 195:

            # Check if shoulders are at the required angle.
            if left_shoulder_angle > 80 and left_shoulder_angle < 110 and right_shoulder_angle > 80 and right_shoulder_angle < 110:

                # Check if it is the warrior II pose.
                # ----------------------------------------------------------------------------------------------------------------

                # Check if one leg is straight.
                if left_knee_angle > 165 and left_knee_angle < 195 or right_knee_angle > 165 and right_knee_angle < 195:

                    # Check if the other leg is bended at the required angle.
                    if left_knee_angle > 90 and left_knee_angle < 120 or right_knee_angle > 90 and right_knee_angle < 120:
                        # Specify the label of the pose that is Warrior II pose.
                        label = 'Warrior II Pose'

                        # ----------------------------------------------------------------------------------------------------------------

                # Check if it is the T pose.
                # ----------------------------------------------------------------------------------------------------------------

                # Check if both legs are straight
                if left_knee_angle > 160 and left_knee_angle < 195 and right_knee_angle > 160 and right_knee_angle < 195:
                    # Specify the label of the pose that is tree pose.
                    label = 'T Pose'

        # ----------------------------------------------------------------------------------------------------------------

        # Check if it is the tree pose.
        # ----------------------------------------------------------------------------------------------------------------

        # Check if one leg is straight
        if left_knee_angle > 165 and left_knee_angle < 195 or right_knee_angle > 165 and right_knee_angle < 195:

            # Check if the other leg is bended at the required angle.
            if left_knee_angle > 315 and left_knee_angle < 335 or right_knee_angle > 25 and right_knee_angle < 45:
                # Specify the label of the pose that is tree pose.
                label = 'Tree Pose'

        # ----------------------------------------------------------------------------------------------------------------

        # Check if the pose is classified successfully
        if label != 'Unknown Pose':
            # Update the color (to green) with which the label will be written on the image.
            color = (255, 255, 0)

            # Write the label on the output image.
        cv2.putText(output_image, label, (70, 250), font_2, 7.8, color, 7)

        # Check if the resultant image is specified to be displayed.
        if display:

            # Display the resultant image.
            plt.figure(figsize=[10, 10])
            plt.imshow(output_image[:, :, ::-1]);
            plt.title("Output Image");
            plt.axis('off');

        else:

            # Return the output image and the classified label.
            return output_image, label

    def calculate_angle(a, b, c):
        a = np.array(a)  # First
        b = np.array(b)  # Mid
        c = np.array(c)  # End

        radians = np.arctan2(c[1] - b[1], c[0] - b[0]) - np.arctan2(a[1] - b[1], a[0] - b[0])
        angle = np.abs(radians * 180.0 / np.pi)

        if angle > 180.0:
            angle = 360 - angle

        return angle

    def modify_AI_elbow(image, landmarks):

        #####  PARAMETER  ######
        width, height = 2048, 1536  ## setting for Full  HD (3K - Max.)
        font_2 = cv2.FONT_HERSHEY_COMPLEX
        AI_elbow = 'This is an A.I. elbow trainer using webcam'
        #####  PARAMETER  ######

        # Get the angle between the left shoulder, elbow and wrist points.
        Left_shoulder = [landmarks[mp_pose.PoseLandmark.LEFT_SHOULDER.value].x,
                         landmarks[mp_pose.PoseLandmark.LEFT_SHOULDER.value].y]
        Right_shoulder = [landmarks[mp_pose.PoseLandmark.RIGHT_SHOULDER.value].x,
                          landmarks[mp_pose.PoseLandmark.RIGHT_SHOULDER.value].y]

        Left_elbow = [landmarks[mp_pose.PoseLandmark.LEFT_ELBOW.value].x,
                      landmarks[mp_pose.PoseLandmark.LEFT_ELBOW.value].y]
        Right_elbow = [landmarks[mp_pose.PoseLandmark.RIGHT_ELBOW.value].x,
                       landmarks[mp_pose.PoseLandmark.RIGHT_ELBOW.value].y]

        Left_wrist = [landmarks[mp_pose.PoseLandmark.LEFT_WRIST.value].x,
                      landmarks[mp_pose.PoseLandmark.LEFT_WRIST.value].y]
        Right_wrist = [landmarks[mp_pose.PoseLandmark.RIGHT_WRIST.value].x,
                       landmarks[mp_pose.PoseLandmark.RIGHT_WRIST.value].y]

        # Calculate elbow angles
        Left_angle = round(calculate_angle(Left_shoulder, Left_elbow, Left_wrist), 2)
        Right_angle = round(calculate_angle(Right_shoulder, Right_elbow, Right_wrist), 2)

        # Visualize elbow angles
        cv2.putText(image, str(Left_angle), tuple(np.multiply(Left_elbow, [width, height]).astype(int)),
                    cv2.FONT_HERSHEY_SIMPLEX, 1.6, (255, 0, 0), 3, cv2.LINE_AA)
        cv2.putText(image, str(Right_angle), tuple(np.multiply(Right_elbow, [width, height]).astype(int)),
                    cv2.FONT_HERSHEY_SIMPLEX, 1.6, (0, 255, 255), 3, cv2.LINE_AA)

        # # Curl counter logic - elbow
        # color = (255, 0, 0)
        # if Left_angle and Right_angle > 160:
        #     stage = "down"
        #     color = (0, 0, 255)
        # if Left_angle and Right_angle < 30 and stage == 'down':
        #     stage = "up"
        #     color = (0, 255, 255)
        #     counter += 1
        #     # print(counter)
        #     speak100(str(counter))

        cv2.putText(image, AI_elbow, (80, 1230), font_2, 2.4, (0, 0, 255), 3, )
        return image, Left_angle, Right_angle

    def modify_AI_shoulder(image, landmarks):

        #####  PARAMETER  ######
        width, height = 2048, 1536  ## setting for Full  HD (3K - Max.)
        font_2 = cv2.FONT_HERSHEY_COMPLEX
        AI_shoulder = 'This is an A.I. shoulder trainer using webcam'
        #####  PARAMETER  ######

        # Get the angle between the  elbow, shoulder and hip points.
        Left_elbow = [landmarks[mp_pose.PoseLandmark.LEFT_ELBOW.value].x,
                      landmarks[mp_pose.PoseLandmark.LEFT_ELBOW.value].y]
        Right_elbow = [landmarks[mp_pose.PoseLandmark.RIGHT_ELBOW.value].x,
                       landmarks[mp_pose.PoseLandmark.RIGHT_ELBOW.value].y]

        Left_shoulder = [landmarks[mp_pose.PoseLandmark.LEFT_SHOULDER.value].x,
                         landmarks[mp_pose.PoseLandmark.LEFT_SHOULDER.value].y]
        Right_shoulder = [landmarks[mp_pose.PoseLandmark.RIGHT_SHOULDER.value].x,
                          landmarks[mp_pose.PoseLandmark.RIGHT_SHOULDER.value].y]

        Left_hip = [landmarks[mp_pose.PoseLandmark.LEFT_HIP.value].x,
                    landmarks[mp_pose.PoseLandmark.LEFT_HIP.value].y]
        Right_hip = [landmarks[mp_pose.PoseLandmark.RIGHT_HIP.value].x,
                     landmarks[mp_pose.PoseLandmark.RIGHT_HIP.value].y]

        # Calculate shoulder angles
        Left_angle = round(calculate_angle(Left_elbow, Left_shoulder, Left_hip), 2)
        Right_angle = round(calculate_angle(Right_elbow, Right_shoulder, Right_hip), 2)

        # Visualize shoulder angles
        cv2.putText(image, str(Left_angle), tuple(np.multiply(Left_shoulder, [width, height]).astype(int)),
                    cv2.FONT_HERSHEY_SIMPLEX, 1.5, (0, 0, 255), 3, cv2.LINE_AA)
        cv2.putText(image, str(Right_angle), tuple(np.multiply(Right_shoulder, [width, height]).astype(int)),
                    cv2.FONT_HERSHEY_SIMPLEX, 1.5, (0, 255, 255), 3, cv2.LINE_AA)

        # cv2.putText(image, AI_shoulder, (80, 1230), font_2, 2.3, (190, 90, 0), 3, )
        cv2.putText(image, AI_shoulder, (80, 1230), font_2, 2.3, (255, 0, 0), 3, )

        return image, Left_angle, Right_angle

    def modify_AI_knee(image, landmarks):

        #####  PARAMETER  ######
        width, height = 2048, 1536  ## setting for Full  HD (3K - Max.)
        font_2 = cv2.FONT_HERSHEY_COMPLEX
        AI_knee = 'This is an A.I. knee trainer using webcam'
        #####  PARAMETER  ######

        # Get the angle between the left hip, knee and ankle points.
        Left_hip = [landmarks[mp_pose.PoseLandmark.LEFT_HIP.value].x,
                    landmarks[mp_pose.PoseLandmark.LEFT_HIP.value].y]
        Right_hip = [landmarks[mp_pose.PoseLandmark.RIGHT_HIP.value].x,
                     landmarks[mp_pose.PoseLandmark.RIGHT_HIP.value].y]

        Left_knee = [landmarks[mp_pose.PoseLandmark.LEFT_KNEE.value].x,
                     landmarks[mp_pose.PoseLandmark.LEFT_KNEE.value].y]
        Right_knee = [landmarks[mp_pose.PoseLandmark.RIGHT_KNEE.value].x,
                      landmarks[mp_pose.PoseLandmark.RIGHT_KNEE.value].y]

        Left_ankle = [landmarks[mp_pose.PoseLandmark.LEFT_ANKLE.value].x,
                      landmarks[mp_pose.PoseLandmark.LEFT_ANKLE.value].y]
        Right_ankle = [landmarks[mp_pose.PoseLandmark.RIGHT_ANKLE.value].x,
                       landmarks[mp_pose.PoseLandmark.RIGHT_ANKLE.value].y]

        # Calculate knee angles
        Left_angle = round(calculate_angle(Left_hip, Left_knee, Left_ankle), 2)
        Right_angle = round(calculate_angle(Right_hip, Right_knee, Right_ankle), 2)

        # Visualize knee angles
        cv2.putText(image, str(Left_angle), tuple(np.multiply(Left_knee, [width, height]).astype(int)),
                    cv2.FONT_HERSHEY_SIMPLEX, 1.5, (0, 255, 255), 3, cv2.LINE_AA)
        cv2.putText(image, str(Right_angle), tuple(np.multiply(Right_knee, [width, height]).astype(int)),
                    cv2.FONT_HERSHEY_SIMPLEX, 1.5, (255, 0, 255), 3, cv2.LINE_AA)

        # cv2.putText(image, AI_knee, (80, 1230), font_2, 2.3, (190, 90, 0), 3, )
        cv2.putText(image, AI_knee, (80, 1230), font_2, 2.3, (255, 88, 255), 3, )

        return image, Left_angle, Right_angle

    def detectPose_surferGame(image, pose, draw=False, display=False):
        '''
        This function performs the pose detection on the most prominent person in an image.
        Args:
            image:   The input image with a prominent person whose pose landmarks needs to be detected.
            pose:    The pose function required to perform the pose detection.
            draw:    A boolean value that is if set to true the function draw pose landmarks on the output image.
            display: A boolean value that is if set to true the function displays the original input image, and the
                     resultant image and returns nothing.
        Returns:
            output_image: The input image with the detected pose landmarks drawn if it was specified.
            results:      The output of the pose landmarks detection on the input image.
        '''

        # Create a copy of the input image.
        output_image = image.copy()

        # Convert the image from BGR into RGB format.
        imageRGB = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)

        # Perform the Pose Detection.
        results = pose.process(imageRGB)

        # Check if any landmarks are detected and are specified to be drawn.
        if results.pose_landmarks and draw:
            # Draw Pose Landmarks on the output image.
            mp_drawing.draw_landmarks(image=output_image,
                                      landmark_list=results.pose_landmarks,
                                      connections=mp_pose.POSE_CONNECTIONS,
                                      landmark_drawing_spec=mp_drawing.DrawingSpec(color=(255, 0, 0), thickness=5,
                                                                                   circle_radius=8),
                                      # landmark_drawing_spec=mp_drawing_styles.get_default_pose_landmarks_style(),  # It show colorful landmarks.
                                      connection_drawing_spec=mp_drawing.DrawingSpec(color=(49, 125, 237), thickness=5,
                                                                                     circle_radius=2)
                                      )

        # Check if the original input image and the resultant image are specified to be displayed.
        if display:

            # Display the original input image and the resultant image.
            plt.figure(figsize=[22, 22])
            plt.subplot(121);
            plt.imshow(image[:, :, ::-1]);
            plt.title("Original Image");
            plt.axis('off');
            plt.subplot(122);
            plt.imshow(output_image[:, :, ::-1]);
            plt.title("Output Image");
            plt.axis('off');

        # Otherwise
        else:

            # Return the output image and the results of pose landmarks detection.
            return output_image, results

    def checkHandsJoined(image, results, draw=False, display=False):
        '''
        This function checks whether the hands of the person are joined or not in an image.
        Args:
            image:   The input image with a prominent person whose hands status (joined or not) needs to be classified.
            results: The output of the pose landmarks detection on the input image.
            draw:    A boolean value that is if set to true the function writes the hands status & distance on the output image.
            display: A boolean value that is if set to true the function displays the resultant image and returns nothing.
        Returns:
            output_image: The same input image but with the classified hands status written, if it was specified.
            hand_status:  The classified status of the hands whether they are joined or not.
        '''

        # Get the height and width of the input image.
        height, width, _ = image.shape

        # Create a copy of the input image to write the hands status label on.
        output_image = image.copy()

        # Get the left wrist landmark x and y coordinates.
        left_wrist_landmark = (results.pose_landmarks.landmark[mp_pose.PoseLandmark.LEFT_WRIST].x * width,
                               results.pose_landmarks.landmark[mp_pose.PoseLandmark.LEFT_WRIST].y * height)

        # Get the right wrist landmark x and y coordinates.
        right_wrist_landmark = (results.pose_landmarks.landmark[mp_pose.PoseLandmark.RIGHT_WRIST].x * width,
                                results.pose_landmarks.landmark[mp_pose.PoseLandmark.RIGHT_WRIST].y * height)

        # Calculate the euclidean distance between the left and right wrist.
        euclidean_distance = int(hypot(left_wrist_landmark[0] - right_wrist_landmark[0],
                                       left_wrist_landmark[1] - right_wrist_landmark[1]))

        # Compare the distance between the wrists with a appropriate threshold to check if both hands are joined.
        if euclidean_distance < 120:

            # Set the hands status to joined.
            hand_status = 'Hands Joined'

            # Set the color value to green.
            color = (255, 255, 0)

        # Otherwise.
        else:

            # Set the hands status to not joined.
            hand_status = 'Hands Not Joined'

            # Set the color value to red.
            color = (255, 0, 255)

        # Check if the Hands Joined status and hands distance are specified to be written on the output image.
        if draw:
            # Write the classified hands status on the image.
            cv2.putText(output_image, hand_status, (70, 100), cv2.FONT_HERSHEY_COMPLEX, 3.0, color, 5)

            # Write the the distance between the wrists on the image.
            cv2.putText(output_image, f'Distance: {euclidean_distance}', (70, 200), cv2.FONT_HERSHEY_COMPLEX, 3.0,
                        color, 5)

        # Check if the output image is specified to be displayed.
        if display:

            # Display the output image.
            plt.figure(figsize=[10, 10]);
            plt.imshow(output_image[:, :, ::-1]);
            plt.title("Output Image");
            plt.axis('off');

        # Otherwise
        else:

            # Return the output image and the classified hands status indicating whether the hands are joined or not.
            return output_image, hand_status

    def checkLeftRight(image, results, draw=False, display=False):
        '''
        This function finds the horizontal position (left, center, right) of the person in an image.
        Args:
            image:   The input image with a prominent person whose the horizontal position needs to be found.
            results: The output of the pose landmarks detection on the input image.
            draw:    A boolean value that is if set to true the function writes the horizontal position on the output image.
            display: A boolean value that is if set to true the function displays the resultant image and returns nothing.
        Returns:
            output_image:         The same input image but with the horizontal position written, if it was specified.
            horizontal_position:  The horizontal position (left, center, right) of the person in the input image.
        '''

        # Declare a variable to store the horizontal position (left, center, right) of the person.
        horizontal_position = None

        # Get the height and width of the image.
        height, width, _ = image.shape

        # Create a copy of the input image to write the horizontal position on.
        output_image = image.copy()

        # Retreive the x-coordinate of the left shoulder landmark.
        left_x = int(results.pose_landmarks.landmark[mp_pose.PoseLandmark.RIGHT_SHOULDER].x * width)

        # Retreive the x-corrdinate of the right shoulder landmark.
        right_x = int(results.pose_landmarks.landmark[mp_pose.PoseLandmark.LEFT_SHOULDER].x * width)

        # Check if the person is at left that is when both shoulder landmarks x-corrdinates
        # are less than or equal to the x-corrdinate of the center of the image.
        if (right_x <= width // 2 and left_x <= width // 2):

            # Set the person's position to left.
            horizontal_position = 'Left'
            # left_sound.play()


        # Check if the person is at right that is when both shoulder landmarks x-corrdinates
        # are greater than or equal to the x-corrdinate of the center of the image.
        elif (right_x >= width // 2 and left_x >= width // 2):

            # Set the person's position to right.
            horizontal_position = 'Right'
            # right_sound.play()

        # Check if the person is at center that is when right shoulder landmark x-corrdinate is greater than or equal to
        # and left shoulder landmark x-corrdinate is less than or equal to the x-corrdinate of the center of the image.
        elif (right_x >= width // 2 and left_x <= width // 2):

            # Set the person's position to center.
            horizontal_position = 'Center'

        # Check if the person's horizontal position and a line at the center of the image is specified to be drawn.
        if draw:
            # Write the horizontal position of the person on the image.
            cv2.putText(output_image, horizontal_position, (25, height - 400), cv2.FONT_HERSHEY_COMPLEX, 4.8,
                        (255, 0, 0), 5)

            # Draw a line at the center of the image.
            cv2.line(output_image, (width // 2, 0), (width // 2, height), (255, 255, 0), 8)

        # Check if the output image is specified to be displayed.
        if display:

            # Display the output image.
            plt.figure(figsize=[10, 10])
            plt.imshow(output_image[:, :, ::-1]);
            plt.title("Output Image");
            plt.axis('off');

        # Otherwise
        else:

            # Return the output image and the person's horizontal position.
            return output_image, horizontal_position

    def checkJumpCrouch(image, results, MID_Y=250, draw=False, display=False):
        '''
        This function checks the posture (Jumping, Crouching or Standing) of the person in an image.
        Args:
            image:   The input image with a prominent person whose the posture needs to be checked.
            results: The output of the pose landmarks detection on the input image.
            MID_Y:   The intial center y-coordinate of both shoulders landmarks of the person recorded during starting
                     the game. This will give the idea of the person's height when he is standing straight.
            draw:    A boolean value that is if set to true the function writes the posture on the output image.
            display: A boolean value that is if set to true the function displays the resultant image and returns nothing.
        Returns:
            output_image: The input image with the person's posture written, if it was specified.
            posture:      The posture (Jumping, Crouching or Standing) of the person in an image.
        '''

        # Get the height and width of the image.
        height, width, _ = image.shape

        # Create a copy of the input image to write the posture label on.
        output_image = image.copy()

        # Retreive the y-coordinate of the left shoulder landmark.
        left_y = int(results.pose_landmarks.landmark[mp_pose.PoseLandmark.RIGHT_SHOULDER].y * height)

        # Retreive the y-coordinate of the right shoulder landmark.
        right_y = int(results.pose_landmarks.landmark[mp_pose.PoseLandmark.LEFT_SHOULDER].y * height)

        # Calculate the y-coordinate of the mid-point of both shoulders.
        actual_mid_y = (abs(right_y + left_y) // 2) - 90  # different between actual_mid_y & lower_bound = 10 (standing)
        # actual_mid_y = (abs(right_y + left_y) // 2) -100 # different between actual_mid_y & lower_bound = 0
        # actual_mid_y = (abs(right_y + left_y) // 2) -110 # different between actual_mid_y & lower_bound = -10 (jumping)

        # print(str(actual_mid_y))  # answer  varies each time.

        # Calculate the upper and lower bounds of the threshold.
        lower_bound = MID_Y - 100
        upper_bound = MID_Y + 50

        # print(str(lower_bound))  #  height of LB & UB = 150
        # print(str(upper_bound))  #  height of LB & UB = 150

        counter = 0

        # Check if the person has jumped that is when the y-coordinate of the mid-point
        # of both shoulders is less than the lower bound.
        if (actual_mid_y < lower_bound):

            counter = 0

            # Set the posture to jumping.
            posture = 'Jumping'
            counter += 1

        # Check if the person has crouched that is when the y-coordinate of the mid-point
        # of both shoulders is greater than the upper bound.
        elif (actual_mid_y > upper_bound):

            # Set the posture to crouching.
            posture = 'Crouching'
            counter += 2

        # Otherwise the person is standing and the y-coordinate of the mid-point
        # of both shoulders is between the upper and lower bounds.
        else:

            # Set the posture to Standing straight.
            posture = 'Standing'

        # Check if the posture and a horizontal line at the threshold is specified to be drawn.
        if draw:
            # Write the posture of the person on the image.
            cv2.putText(output_image, posture, (25, height - 125), cv2.FONT_HERSHEY_COMPLEX, 4.8, (255, 0, 0), 5)
            # cv2.putText(output_image, str(counter), (790, height - 135), cv2.FONT_HERSHEY_COMPLEX, 4.8, (255, 0, 0), 7)

            # Draw a line at the intial center y-coordinate of the person (threshold).
            cv2.line(output_image, (0, MID_Y), (width, MID_Y), (255, 255, 0), 12)

        # Check if the output image is specified to be displayed.
        if display:

            # Display the output image.
            plt.figure(figsize=[10, 10])
            plt.imshow(output_image[:, :, ::-1]);
            plt.title("Output Image");
            plt.axis('off');

        # Otherwise
        else:

            # Return the output image and posture indicating whether the person is standing straight or has jumped, or crouched.
            return output_image, posture


def main():
    cap = cv2.VideoCapture(2)
    detector = FaceDetector()

    while True:
        success, img = cap.read()
        img, bboxs = detector.findFaces(img)
        # print(bboxs)
        cv2.imshow("Image", img)
        key = cv2.waitKey(1)
        if key == 27:
            break
        if key == ord('q'):
            break

if __name__ == "__main__":
    main()



"""
Hand Tracing Module
"""

import cv2
import mediapipe as mp
import os, time
import numpy as np
from colorama import Fore
import matplotlib.pyplot as plt
import subprocess
import pyautogui as pg
import pyglet
import math



# Initialize the mediapipe hands class.
mp_hands = mp.solutions.hands
# Initialize the mediapipe drawing class.
mp_drawing = mp.solutions.drawing_utils
# It show colorful hands landmarks
mp_drawing_styles = mp.solutions.drawing_styles

# Set up the Hands functions for images.
# hands = mp_hands.Hands(static_image_mode=True, max_num_hands=2, min_detection_confidence=0.75) # ## For images only.

# Set up the Hands functions for  videos.
hands_videos = mp_hands.Hands(static_image_mode=False, max_num_hands=2,
                              min_detection_confidence=0.75, min_tracking_confidence=0.75)



############ set the size of web cam ########q#####
# width, height = 640, 480  ## setting for default
# width, height = 1280, 720  ## setting for HD
width, height = 1920, 1080  ## setting for Full  HD (2K)
# width, height =3840, 2160  ## setting for 4K
 ############ set the size of web cam ########q#####


class handDetector():

    def __init__(self, mode=False, maxHands=2, detectionCon=0.5, trackCon=0.5):
        self.mode = mode
        self.maxHands = maxHands
        self.detectionCon = detectionCon
        self.trackCon = trackCon

        self.mpHands = mp.solutions.hands
        self.hands = self.mpHands.Hands(self.mode, self.maxHands,
                                        self.detectionCon, self.trackCon)
        self.mpDraw = mp.solutions.drawing_utils
        self.tipIds = [4, 8, 12, 16, 20]

    def findHands(self, img, draw=True):
            imgRGB = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
            self.results = self.hands.process(imgRGB)
            # print(results.multi_hand_landmarks)

            if self.results.multi_hand_landmarks:
                for handLms in self.results.multi_hand_landmarks:
                    if draw:
                        self.mpDraw.draw_landmarks(img, handLms,self.mpHands.HAND_CONNECTIONS)
            return img

    def findPosition(self, img, handNo=0, draw=True):
        xList = []
        yList = []
        bbox = []
        self.lmList = []
        if self.results.multi_hand_landmarks:
            myHand = self.results.multi_hand_landmarks[handNo]
            for id, lm in enumerate(myHand.landmark):
                # print(id, lm)
                h, w, c = img.shape
                cx, cy = int(lm.x * w), int(lm.y * h)
                xList.append(cx)
                yList.append(cy)
                # print(id, cx, cy)
                self.lmList.append([id, cx, cy])
                if draw:
                    cv2.circle(img, (cx, cy), 5, (255, 0, 255), cv2.FILLED)
            xmin, xmax = min(xList), max(xList)
            ymin, ymax = min(yList), max(yList)
            bbox = xmin, ymin, xmax, ymax

            if draw:
                cv2.rectangle(img, (bbox[0] - 20, bbox[1] - 20),
                              (bbox[2] + 20, bbox[3] + 20), (0, 255, 0), 2)

        return self.lmList, bbox

    def fingersUp(self):
        fingers = []
        # Thumb
        if self.lmList[self.tipIds[0]][1] < self.lmList[self.tipIds[0] - 1][1]:
            fingers.append(1)
        else:
            fingers.append(0)

        # 4 Fingers
        for id in range(1, 5):
            if self.lmList[self.tipIds[id]][2] < self.lmList[self.tipIds[id] - 2][2]:
                fingers.append(1)
            else:
                fingers.append(0)
        return fingers

    def findDistance(self, p1, p2, img, draw=True):

        x1, y1 = self.lmList[p1][1], self.lmList[p1][2]
        x2, y2 = self.lmList[p2][1], self.lmList[p2][2]
        cx, cy = (x1 + x2) // 2, (y1 + y2) // 2

        if draw:
            cv2.circle(img, (x1, y1), 15, (255, 0, 255), cv2.FILLED)
            cv2.circle(img, (x2, y2), 15, (255, 0, 255), cv2.FILLED)
            cv2.line(img, (x1, y1), (x2, y2), (255, 0, 255), 3)
            cv2.circle(img, (cx, cy), 15, (255, 0, 255), cv2.FILLED)

        length = math.hypot(x2 - x1, y2 - y1)
        return length, img, [x1, y1, x2, y2, cx, cy]

    def get_label(self,index, hand, results):
        output = None
        for idx, classification in enumerate(results.multi_handedness):
            if classification.classification[0].index == index:
                label = classification.classification[0].label
                # score = classification.classification[0].score
                text = '{}'.format(label)
                # text = '{} {}'.format(label, round(score, 3))
                coords = tuple(np.multiply(np.array((hand.landmark[mp_hands.HandLandmark.WRIST].x,
                                                     hand.landmark[mp_hands.HandLandmark.WRIST].y)),
                                           [width, height]).astype(int))
                output = text, coords
                return output

    def draw_finger_angles(self,image, results, joint_list):
        # Load sounds
        sound = pyglet.media.load("sound.wav", streaming=False)

        # Loop through hands
        for hand in results.multi_hand_landmarks:

            # Loop through joint sets
            for joint in joint_list:
                a = np.array([hand.landmark[joint[0]].x, hand.landmark[joint[0]].y])  # First coord
                b = np.array([hand.landmark[joint[1]].x, hand.landmark[joint[1]].y])  # Second coord
                c = np.array([hand.landmark[joint[2]].x, hand.landmark[joint[2]].y])  # Third coord

                radians = np.arctan2(c[1] - b[1], c[0] - b[0]) - np.arctan2(a[1] - b[1], a[0] - b[0])
                angle = np.abs(radians * 180.0 / np.pi)
                if angle > 180.0:
                    angle = 360 - angle
                    cv2.putText(image, str(round(angle, 1)), tuple(np.multiply(b, [width, height]).astype(int)),
                                cv2.FONT_HERSHEY_SIMPLEX, (0.5 * 1.8), (255, 200, 200), (1 * 2), cv2.LINE_AA)

                ######### 1. bending thumb --  open the scientific calculator ####################
                if angle < 150.0 and joint==joint_list[0] :
                    sound.play()
                    speak100("let's open the scientific calculator for you sir.")
                    subprocess.Popen('C:/Windows/System32/calc.exe')
                ######### 1. bending thumb -- open the scientific calculator ####################

                ######### 2. bending index finger -- open the video of Bruce Lee ####################
                if angle < 150.0 and joint == joint_list[1]:
                    # speak100("Oops!")
                    sound.play()
                    speak100("the video of Bruce Lee versus Donnie Yen in : a warriors dream ")
                #     os.startfile("https://www.youtube.com/watch?v=Gxtvn23__z4")
                #     # https://www.youtube.com/watch?v=Gxtvn23__z4
                #     time.sleep(5)  ## It can not < 5
                #     pg.click(x=770, y=540, interval=1)
                #     # pg.click(x=1202, y=775, interval=1)
                ######### 2.  bending index finger -- open the video of Bruce Lee ####################

                ########## 3. bending middle finger -- open Sinsuran file  ####################
                if angle < 150.0 and joint == joint_list[2] :
                    sound.play()
                    # speak100("Oops!")
                    speak100("Let's open the LibreOffice calc file of  in Sinsuran, Kota Kinabalu. ")
                    # os.startfile("C:/Users/acer/Desktop/Sinsuran/Sinsuran 2021/Sinsuran - 2 Day payment "
                    #              "- V21- 1Jan2021.ods")
                ######### 3. bending middle finger --  open Sinsuran file  ####################

                ########## 4.  bending ring finger  --  open the The Daily Express newspaper ####################
                if angle < 150.0 and joint == joint_list[3]:
                    sound.play()
                    # speak100("Oops!")
                #     pg.FAILSAFE = True
                    speak100("let's open The Daily Express newspaper for you sir.")
                #     pg.hotkey("winleft")
                #     time.sleep(3)
                #     pg.click(x=833, y=630, interval=1)
                #     # in pyAutoGui the typewrite command the \n means hit enter & not new line.
                #     pg.typewrite('www.dailyexpress.com.my\n', 0.2)
                ######### 4.  bending ring finger  --  open the The Daily Express newspaper ####################

                ######### 5. bending pinky finger --  play the song by Pet Shop Boys & take photos ####################
                if angle < 150.0 and joint == joint_list[4]:
                    sound.play()
                    # speak100("Oops!")
                    speak100("let's play the song by Pet Shop Boys - Domino Dancing. ")
                #     os.startfile('C:/Users/acer/Desktop/songs/Pet Shop Boys - Domino Dancing.mp3')
                # ##  or ##
                # if angle < 150.0 and joint == joint_list[4]:
                    # pg.FAILSAFE = True
                    # speak100("let's open the camera and take picture for you sir.")
                    # pg.hotkey("winleft")
                    # time.sleep(2)
                    # pg.click(x=670, y=460, interval=0.25)  ##click main window
                    # # time.sleep(3)
                    # # pg.click(x=1257, y=(485 + 170), interval=0.25)  ##click camera icon
                    # time.sleep(5)
                    # pg.click(x=1265, y=495, interval=0.25)  ##click start camera button
                    # # pg.click(x=1257, y=485, interval=0.25)
                ######### 5. bending pinky finger --  play the song by Pet Shop Boys  & take photos ####################
            return image

    def make_folder(self, myFolder):
        if not os.path.exists(myFolder):
            os.makedirs(myFolder)
            print('Creating subfolder: Tim_folder')
            print('@============ A new sub-folder created for you, sir! =================@')
        return myFolder

#
    def detectHandsLandmarks(image, hands_videos, draw=True, display=True):
        '''
        This function performs hands landmarks detection on an image.
        Args:
            image:   The input image with prominent hand(s) whose landmarks needs to be detected.
            hands:   The Hands function required to perform the hands landmarks detection.
            draw:    A boolean value that is if set to true the function draws hands landmarks on the output image.
            display: A boolean value that is if set to true the function displays the original input image, and the output
                     image with hands landmarks drawn if it was specified and returns nothing.
        Returns:
            output_image: A copy of input image with the detected hands landmarks drawn if it was specified.
            results:      The output of the hands landmarks detection on the input image.
        '''

        # Create a copy of the input image to draw landmarks on.
        output_image = image.copy()

        # Convert the image from BGR into RGB format.
        imgRGB = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)

        # Perform the Hands Landmarks Detection.
        results = hands_videos.process(imgRGB)

        # Check if landmarks are found and are specified to be drawn.
        if results.multi_hand_landmarks and draw:

            # Iterate over the found hands.
            for hand_landmarks in results.multi_hand_landmarks:
                # Draw the hand landmarks on the copy of the input image.
                mp_drawing.draw_landmarks(image=output_image, landmark_list=hand_landmarks,
                                          connections=mp_hands.HAND_CONNECTIONS,
                                          landmark_drawing_spec=mp_drawing_styles.get_default_hand_landmarks_style(),
                                          ## Alternative way.
                                          connection_drawing_spec=mp_drawing_styles.get_default_hand_connections_style())  ## Alternative way.
                # landmark_drawing_spec=mp_drawing.DrawingSpec(color=(255 ,255 ,255),thickness=5, circle_radius=5),
                # connection_drawing_spec=mp_drawing.DrawingSpec(color=(0 ,255 ,0), thickness=5, circle_radius=7))

        # Check if the original input image and the output image are specified to be displayed.
        if display:

            # Display the original input image and the output image.
            plt.figure(figsize=[15, 15])
            plt.subplot(121) \
                ;
            plt.imshow(image[:, :, ::-1]) \
                ;
            plt.title("Original Image") \
                ;
            plt.axis('off');
            plt.subplot(122) \
                ;
            plt.imshow(output_image[:, :, ::-1]) \
                ;
            plt.title("Output") \
                ;
            plt.axis('off');

        # Otherwise
        else:

            # Return the output image and results of hands landmarks detection.
            return output_image, results

        # Now let's test the function **`detectHandsLandmarks()`** created above to perform hands landmarks detection on a sample image and display the results.

    def countFingers(image, results, draw=True, display=True):
        '''
        This function will count the number of fingers up for each hand in the image.
        Args:
            image:   The image of the hands on which the fingers counting is required to be performed.
            results: The output of the hands landmarks detection performed on the image of the hands.
            draw:    A boolean value that is if set to true the function writes the total count of fingers of the hands on the
                     output image.
            display: A boolean value that is if set to true the function displays the resultant image and returns nothing.
        Returns:
            output_image:     A copy of the input image with the fingers count written, if it was specified.
            fingers_statuses: A dictionary containing the status (i.e., open or close) of each finger of both hands.
            count:            A dictionary containing the count of the fingers that are up, of both hands.
        '''

        # Get the height and width of the input image.
        height, width, _ = image.shape

        # Create a copy of the input image to write the count of fingers on.
        output_image = image.copy()

        # Initialize a dictionary to store the count of fingers of both hands.
        count = {'RIGHT': 0, 'LEFT': 0}

        # Store the indexes of the tips landmarks of each finger of a hand in a list.
        fingers_tips_ids = [mp_hands.HandLandmark.INDEX_FINGER_TIP, mp_hands.HandLandmark.MIDDLE_FINGER_TIP,
                            mp_hands.HandLandmark.RING_FINGER_TIP, mp_hands.HandLandmark.PINKY_TIP]

        # Initialize a dictionary to store the status (i.e., True for open and False for close) of each finger of both hands.
        fingers_statuses = {'RIGHT_THUMB': False, 'RIGHT_INDEX': False, 'RIGHT_MIDDLE': False, 'RIGHT_RING': False,
                            'RIGHT_PINKY': False, 'LEFT_THUMB': False, 'LEFT_INDEX': False, 'LEFT_MIDDLE': False,
                            'LEFT_RING': False, 'LEFT_PINKY': False}

        # Iterate over the found hands in the image.
        for hand_index, hand_info in enumerate(results.multi_handedness):

            # Retrieve the label of the found hand.
            hand_label = hand_info.classification[0].label

            # Retrieve the landmarks of the found hand.
            hand_landmarks = results.multi_hand_landmarks[hand_index]

            # Iterate over the indexes of the tips landmarks of each finger of the hand.
            for tip_index in fingers_tips_ids:

                # Retrieve the label (i.e., index, middle, etc.) of the finger on which we are iterating upon.
                finger_name = tip_index.name.split("_")[0]

                # Check if the finger is up by comparing the y-coordinates of the tip and pip landmarks.
                if (hand_landmarks.landmark[tip_index].y < hand_landmarks.landmark[tip_index - 2].y):
                    # Update the status of the finger in the dictionary to true.
                    fingers_statuses[hand_label.upper() + "_ " + finger_name] = True

                    # Increment the count of the fingers up of the hand by 1.
                    count[hand_label.upper()] += 1

            # Retrieve the y-coordinates of the tip and mcp landmarks of the thumb of the hand.
            thumb_tip_x = hand_landmarks.landmark[mp_hands.HandLandmark.THUMB_TIP].x
            thumb_mcp_x = hand_landmarks.landmark[mp_hands.HandLandmark.THUMB_TIP - 2].x

            # Check if the thumb is up by comparing the hand label and the x-coordinates of the retrieved landmarks.
            if (hand_label == 'Right' and (thumb_tip_x < thumb_mcp_x)) or (hand_label == 'Left'
                                                                           and (thumb_tip_x > thumb_mcp_x)):
                # Update the status of the thumb in the dictionary to true.
                fingers_statuses[hand_label.upper() + "_THUMB"] = True

                # Increment the count of the fingers up of the hand by 1.
                count[hand_label.upper()] += 1

        # Check if the total count of the fingers of both hands are specified to be written on the output image.
        if draw:
            # Write the total count of the fingers of both hands on the output image.
            cv2.putText(output_image, " Total Fingers: ", (450, 68), cv2.FONT_HERSHEY_COMPLEX, 1.5, (20, 255, 155), 3)
            cv2.putText(output_image, str(sum(count.values())), (width // 2 - 100, 125), cv2.FONT_HERSHEY_SIMPLEX,
                        5.0, (20, 255, 155), 12, 10)

        # Check if the output image is specified to be displayed.
        if display:

            # Display the output image.
            plt.figure(figsize=[10, 10])
            plt.imshow(output_image[:, :, ::-1]);
            plt.title("Output Image");
            plt.axis('off');

        # Otherwise
        else:

            # Return the output image, the status of each finger and the count of the fingers up of both hands.
            return output_image, fingers_statuses, count

    def annotate(image, results, fingers_statuses, count, display=True):
        '''
        This function will draw an appealing visualization of each fingers up of the both hands in the image.
        Args:
            image:            The image of the hands on which the counted fingers are required to be visualized.
            results:          The output of the hands landmarks detection performed on the image of the hands.
            fingers_statuses: A dictionary containing the status (i.e., open or close) of each finger of both hands.
            count:            A dictionary containing the count of the fingers that are up, of both hands.
            display:          A boolean value that is if set to true the function displays the resultant image and
                              returns nothing.
        Returns:
            output_image: A copy of the input image with the visualization of counted fingers.
        '''

        # Get the height and width of the input image.
        height, width, _ = image.shape

        # Create a copy of the input image.
        output_image = image.copy()

        # Select the images of the hands prints that are required to be overlayed.
        ########################################################################################################################

        # Initialize a dictionaty to store the images paths of the both hands.
        # Initially it contains red hands images paths. The red image represents that the hand is not present in the image.
        HANDS_IMGS_PATHS = {'LEFT': ['media/left_hand_not_detected.png'],
                            'RIGHT': ['media/right_hand_not_detected.png']}

        # Check if there is hand(s) in the image.
        if results.multi_hand_landmarks:

            # Iterate over the detected hands in the image.
            for hand_index, hand_info in enumerate(results.multi_handedness):

                # Retrieve the label of the hand.
                hand_label = hand_info.classification[0].label

                # Update the image path of the hand to a green color hand image.
                # This green image represents that the hand is present in the image.
                HANDS_IMGS_PATHS[hand_label.upper()] = ['media/' + hand_label.lower() + '_hand_detected.png']

                # Check if all the fingers of the hand are up/open.
                if count[hand_label.upper()] == 5:

                    # Update the image path of the hand to a hand image with green color palm and orange color fingers image.
                    # The orange color of a finger represents that the finger is up.
                    HANDS_IMGS_PATHS[hand_label.upper()] = ['media/' + hand_label.lower() + '_all_fingers.png']

                # Otherwise if all the fingers of the hand are not up/open.
                else:

                    # Iterate over the fingers statuses of the hands.
                    for finger, status in fingers_statuses.items():

                        # Check if the finger is up and belongs to the hand that we are iterating upon.
                        if status == True and finger.split("_")[0] == hand_label.upper():
                            # Append another image of the hand in the list inside the dictionary.
                            # This image only contains the finger we are iterating upon of the hand in orange color.
                            # As the orange color represents that the finger is up.
                            HANDS_IMGS_PATHS[hand_label.upper()].append('media/' + finger.lower() + '.png')

        ########################################################################################################################

        # Overlay the selected hands prints on the input image.
        ########################################################################################################################

        # Iterate over the left and right hand.
        for hand_index, hand_imgs_paths in enumerate(HANDS_IMGS_PATHS.values()):

            # Iterate over the images paths of the hand.
            for img_path in hand_imgs_paths:
                # Read the image including its alpha channel. The alpha channel (0-255) determine the level of visibility.
                # In alpha channel, 0 represents the transparent area and 255 represents the visible area.
                hand_imageBGRA = cv2.imread(img_path, cv2.IMREAD_UNCHANGED)

                # Retrieve all the alpha channel values of the hand image.
                alpha_channel = hand_imageBGRA[:, :, -1]

                # Retrieve all the blue, green, and red channels values of the hand image.
                # As we also need the three-channel version of the hand image.
                hand_imageBGR = hand_imageBGRA[:, :, :-1]

                # Retrieve the height and width of the hand image.
                hand_height, hand_width, _ = hand_imageBGR.shape

                # Retrieve the region of interest of the output image where the handprint image will be placed.
                ROI = output_image[30: 30 + hand_height,
                      (hand_index * width // 2) + width // 12: ((hand_index * width // 2) + width // 12 + hand_width)]

                # Overlay the handprint image by updating the pixel values of the ROI of the output image at the
                # indexes where the alpha channel has the value 255.
                ROI[alpha_channel == 255] = hand_imageBGR[alpha_channel == 255]

                # Update the ROI of the output image with resultant image pixel values after overlaying the handprint.
                output_image[30: 30 + hand_height,
                (hand_index * width // 2) + width // 12: ((hand_index * width // 2) + width // 12 + hand_width)] = ROI

        ########################################################################################################################

        # Check if the output image is specified to be displayed.
        if display:

            # Display the output image.
            plt.figure(figsize=[10, 10])
            plt.imshow(output_image[:, :, ::-1]);
            plt.title("Output Image");
            plt.axis('off');

        # Otherwise
        else:

            # Return the output image
            return output_image

    def recognizeGestures(image, fingers_statuses, count, draw=True, display=True):
        '''
        This function will determine the gesture of the left and right hand in the image.
        Args:
            image:            The image of the hands on which the hand gesture recognition is required to be performed.
            fingers_statuses: A dictionary containing the status (i.e., open or close) of each finger of both hands.
            count:            A dictionary containing the count of the fingers that are up, of both hands.
            draw:             A boolean value that is if set to true the function writes the gestures of the hands on the
                              output image, after recognition.
            display:          A boolean value that is if set to true the function displays the resultant image and
                              returns nothing.
        Returns:
            output_image:   A copy of the input image with the left and right hand recognized gestures written if it was
                            specified.
            hands_gestures: A dictionary containing the recognized gestures of the right and left hand.
        '''

        # Create a copy of the input image.
        output_image = image.copy()

        # Store the labels of both hands in a list.
        hands_labels = ['RIGHT', 'LEFT']

        # Initialize a dictionary to store the gestures of both hands in the image.
        hands_gestures = {'RIGHT': "UNKNOWN", 'LEFT': "UNKNOWN"}

        # Iterate over the left and right hand.
        for hand_index, hand_label in enumerate(hands_labels):

            # Initialize a variable to store the color we will use to write the hands gestures on the image.
            # Initially it is red which represents that the gesture is not recognized.
            color = (0, 0, 255)

            ####################################################################################################################
            # 1. Check if the person is making the 'V' gesture with the hand.
            ####################################################################################################################

            # Check if the number of fingers up is 2 and the fingers that are up, are the index and the middle finger.
            if count[hand_label] == 2 and fingers_statuses[hand_label + '_MIDDLE'] and fingers_statuses[
                hand_label + '_INDEX']:

                # Update the gesture value of the hand that we are iterating upon to V SIGN.
                hands_gestures[hand_label] = "V SIGN"

                # Update the color value to green.
                color = (0, 255, 0)

            ###############################################################################################################
            # 2. Check if the person is making the 'SPIDERMAN' gesture with the hand.
            ##########################################################################################################################################################

            # Check if the number of fingers up is 3 and the fingers that are up, are the thumb, index and the pinky finger.
            elif count[hand_label] == 3 and fingers_statuses[hand_label + '_THUMB'] and fingers_statuses[
                hand_label + '_INDEX'] and fingers_statuses[hand_label + '_PINKY']:

                # Update the gesture value of the hand that we are iterating upon to SPIDERMAN SIGN.
                hands_gestures[hand_label] = "SPIDERMAN SIGN"

                # Update the color value to green.
                color = (0, 255, 0)

            ###############################################################################################################
            # 3. Check if the person is making the 'HIGH-FIVE' gesture with the hand.
            ####################################################################################################################
            # Check if the number of fingers up is 5, which means that all the fingers are up.
            elif count[hand_label] == 5:

                # Update the gesture value of the hand that we are iterating upon to HIGH-FIVE SIGN.
                hands_gestures[hand_label] = "HIGH-FIVE SIGN"

                # Update the color value to green.
                color = (0, 255, 0)

            ####################################################################################################################
            # 4. Check if the person is making the '1' gesture with the hand.
            ####################################################################################################################
            # Check if the number of fingers up is 1 and the fingers that is up is the index finger.
            elif count[hand_label] == 1 and fingers_statuses[hand_label + '_INDEX']:
                # Update the gesture value of the hand that we are iterating upon to V SIGN.
                hands_gestures[hand_label] = "1 SIGN"

                # Update the color value to green.
                color = (0, 255, 0)

            ##################################################################################################################
            ####################################################################################################################
            # 5. Check if the person is making the '6' gesture with the hand.
            ####################################################################################################################
            # Check if the number of fingers up is 3 and the fingers that are up, are the thumband &the pinky finger.
            elif count[hand_label] == 2 and fingers_statuses[hand_label + '_THUMB'] and fingers_statuses[
                hand_label + '_PINKY']:
                # Update the gesture value of the hand that we are iterating upon to V SIGN.
                hands_gestures[hand_label] = "6 SIGN"

                # Update the color value to green.
                color = (0, 255, 0)


            ##################################################################################################################
            ####################################################################################################################
            # 6. Check if the person is making the 'Good' gesture with the hand.
            ####################################################################################################################
            # Check if the number of fingers up is 1 and the fingers that are up, are the thumband .
            elif count[hand_label] == 1 and fingers_statuses[hand_label + '_THUMB']:

                # Update the gesture value of the hand that we are iterating upon to V SIGN.
                hands_gestures[hand_label] = "Good SIGN"

                # Update the color value to green.
                color = (0, 255, 0)

            ####################################################################################################################
            # 7. Check if the person is making the 'PINKY FINGER SIGN' gesture with the hand.
            ####################################################################################################################
            # Check if the number of fingers up is 1 and the fingers that are up, are the thumband .
            elif count[hand_label] == 1 and fingers_statuses[hand_label + '_PINKY']:

                # Update the gesture value of the hand that we are iterating upon to PINKY FINGER SIGN.
                hands_gestures[hand_label] = "PINKY FINGER SIGN"

                # Update the color value to green.
                color = (0, 255, 0)
            ####################################################################################################################
            # 8. Check if the person is making the '3 FINGER SIGN' gesture with the hand.
            ####################################################################################################################
            # Check if the number of fingers up is 3 and the fingers that are up, are the thumband .
            elif count[hand_label] == 3 and fingers_statuses[hand_label + '_INDEX'] and fingers_statuses[
                hand_label + '_MIDDLE'] and fingers_statuses[hand_label + '_RING']:
                # Update the gesture value of the hand that we are iterating upon to PINKY FINGER SIGN.
                hands_gestures[hand_label] = "3 FINGER SIGN"

                # Update the color value to green.
                color = (0, 255, 0)

            ####################################################################################################################
            # 9. Check if the person is making the '4 FINGER SIGN' gesture with the hand.
            ####################################################################################################################
            # Check if the number of fingers up is 4 and the fingers that are up, are the thumband .
            elif count[hand_label] == 4 and fingers_statuses[hand_label + '_INDEX'] and fingers_statuses[
                hand_label + '_MIDDLE'] and fingers_statuses[hand_label + '_RING'] and fingers_statuses[
                hand_label + '_PINKY']:
                # Update the gesture value of the hand that we are iterating upon to PINKY FINGER SIGN.
                hands_gestures[hand_label] = "4 FINGER SIGN"

                # Update the color value to green.
                color = (0, 255, 0)

            ####################################################################################################################
            # 10. Check if the person is making the 'OK SIGN' gesture with the hand.
            ####################################################################################################################
            # Check if the number of fingers up is 4 and the fingers that are up, are the thumband .
            elif count[hand_label] == 4 and fingers_statuses[hand_label + '_THUMB'] and fingers_statuses[
                hand_label + '_MIDDLE'] and fingers_statuses[hand_label + '_RING'] and fingers_statuses[
                hand_label + '_PINKY']:
                # Update the gesture value of the hand that we are iterating upon to PINKY FINGER SIGN.
                hands_gestures[hand_label] = "OK SIGN"

                # Update the color value to green.
                color = (0, 255, 255)

            ####################################################################################################################
            # 11. Check if the person is making the 'GUN SIGN' gesture with the hand.
            ####################################################################################################################
            # Check if the number of fingers up is 3 and the fingers that are up, are the thumband .
            elif count[hand_label] == 3 and fingers_statuses[hand_label + '_THUMB'] and fingers_statuses[
                hand_label + '_INDEX'] and fingers_statuses[hand_label + '_MIDDLE']:

                # Update the gesture value of the hand that we are iterating upon to PINKY FINGER SIGN.
                hands_gestures[hand_label] = "GUN SIGN"

                # Update the color value to green.
                color = (0, 255, 255)

            ####################################################################################################################
            # 12. Check if the person is making the 'SCOUT SIGN' gesture with the hand.
            ####################################################################################################################
            # Check if the number of fingers up is 3 and the fingers that are up, are the thumband .
            elif count[hand_label] == 3 and fingers_statuses[hand_label + '_THUMB'] and fingers_statuses[
                hand_label + '_MIDDLE'] and fingers_statuses[hand_label + '_RING']:

                # Update the gesture value of the hand that we are iterating upon to PINKY FINGER SIGN.
                hands_gestures[hand_label] = "SCOUT SIGN"

                # Update the color value to green.
                color = (0, 255, 255)

            ####################################################################################################################
            # 13. Check if the person is making the 'MONKEY SIGN' gesture with the hand.
            ####################################################################################################################
            # Check if the number of fingers up is 4 and the fingers that are up, are the thumband .
            elif count[hand_label] == 4 and fingers_statuses[hand_label + '_THUMB'] \
                    and fingers_statuses[hand_label + '_INDEX'] and fingers_statuses[
                hand_label + '_MIDDLE'] and fingers_statuses[hand_label + '_RING']:

                # Update the gesture value of the hand that we are iterating upon to PINKY FINGER SIGN.
                hands_gestures[hand_label] = "MONKEY SIGN"

                # Update the color value to green.
                color = (0, 255, 255)
            ##################################################################################################################
            ##################################################################################################################
            ##################################################################################################################

            # Check if the hands gestures are specified to be written.
            if draw:
                # Write the hand gesture on the output image.
                cv2.putText(output_image, hand_label + ': ' + hands_gestures[hand_label], (10, (hand_index + 1) * 60),
                            cv2.FONT_HERSHEY_PLAIN, 4, color, 5)

        # Check if the output image is specified to be displayed.
        if display:

            # Display the output image.
            plt.figure(figsize=[10, 10])
            plt.imshow(output_image[:, :, ::-1]);
            plt.title("Output Image");
            plt.axis('off');

        # Otherwise
        else:

            # Return the output image and the gestures of the both hands.
            return output_image, hands_gestures

    def getHandType(image, results, draw=True, display=True):
        '''
        This function performs hands type (left or right) classification on hands.
        Args:
            image:   The image of the hands that needs to be classified, with the hands landmarks detection already performed.
            results: The output of the hands landmarks detection performed on the image in which hands types needs
                     to be classified.
            draw:    A boolean value that is if set to true the function writes the hand type label on the output image.
            display: A boolean value that is if set to true the function displays the output image and returns nothing.
        Returns:
            output_image: The image of the hands with the classified hand type label written if it was specified.
            hands_status: A dictionary containing classification info of both hands.
        '''

        # Create a copy of the input image to write hand type label on.
        output_image = image.copy()

        # Initialize a dictionary to store the classification info of both hands.
        hands_status = {'Right': False, 'Left': False, 'Right_index': None, 'Left_index': None}

        # Iterate over the found hands in the image.
        for hand_index, hand_info in enumerate(results.multi_handedness):

            # Retrieve the label of the found hand.
            hand_type = hand_info.classification[0].label

            # Update the status of the found hand.
            hands_status[hand_type] = True

            # Update the index of the found hand.
            hands_status[hand_type + '_index'] = hand_index

            # Check if the hand type label is specified to be written.
            if draw:
                # Write the hand type on the output image.
                cv2.putText(output_image, hand_type + ' Hand Detected', (10, (hand_index + 3) * 60),
                            cv2.FONT_HERSHEY_PLAIN,
                            7, (255, 255, 0), 4)

        # Check if the output image is specified to be displayed.
        if display:

            # Display the output image.
            plt.figure(figsize=[10, 10])
            plt.imshow(output_image[:, :, ::-1]);
            plt.title("Output Image");
            plt.axis('off');

        # Otherwise
        else:

            # Return the output image and the hands status dictionary that contains classification info.
            return output_image, hands_status

    def drawBoundingBoxes(image, results, hand_status, padd_amount=10, draw=True, display=True):
        '''
        This function draws bounding boxes around the hands and write their classified types near them.
        Args:
            image:       The image of the hands on which the bounding boxes around the hands needs to be drawn and the
                         classified hands types labels needs to be written.
            results:     The output of the hands landmarks detection performed on the image on which the bounding boxes needs
                         to be drawn.
            hand_status: The dictionary containing the classification info of both hands.
            padd_amount: The value that specifies the space inside the bounding box between the hand and the box's borders.
            draw:        A boolean value that is if set to true the function draws bounding boxes and write their classified
                         types on the output image.
            display:     A boolean value that is if set to true the function displays the output image and returns nothing.
        Returns:
            output_image:     The image of the hands with the bounding boxes drawn and hands classified types written if it
                              was specified.
            output_landmarks: The dictionary that stores both (left and right) hands landmarks as different elements.
        '''

        # Create a copy of the input image to draw bounding boxes on and write hands types labels.
        output_image = image.copy()

        # Initialize a dictionary to store both (left and right) hands landmarks as different elements.
        output_landmarks = {}

        # Get the height and width of the input image.
        height, width, _ = image.shape

        # Iterate over the found hands.
        for hand_index, hand_landmarks in enumerate(results.multi_hand_landmarks):

            # Initialize a list to store the detected landmarks of the hand.
            landmarks = []

            # Iterate over the detected landmarks of the hand.
            for landmark in hand_landmarks.landmark:
                # Append the landmark into the list.
                landmarks.append((int(landmark.x * width), int(landmark.y * height),
                                  (landmark.z * width)))

            # Get all the x-coordinate values from the found landmarks of the hand.
            x_coordinates = np.array(landmarks)[:, 0]

            # Get all the y-coordinate values from the found landmarks of the hand.
            y_coordinates = np.array(landmarks)[:, 1]

            # Get the bounding box coordinates for the hand with the specified padding.
            x1 = int(np.min(x_coordinates) - padd_amount)
            y1 = int(np.min(y_coordinates) - padd_amount)
            x2 = int(np.max(x_coordinates) + padd_amount)
            y2 = int(np.max(y_coordinates) + padd_amount)

            # Initialize a variable to store the label of the hand.
            label = "Unknown"

            # Check if the hand we are iterating upon is the right one.
            if hand_status['Right_index'] == hand_index:

                # Update the label and store the landmarks of the hand in the dictionary.
                label = 'Right Hand'
                output_landmarks['Right'] = landmarks

            # Check if the hand we are iterating upon is the left one.
            elif hand_status['Left_index'] == hand_index:

                # Update the label and store the landmarks of the hand in the dictionary.
                label = 'Left Hand'
                output_landmarks['Left'] = landmarks

            # Check if the bounding box and the classified label is specified to be written.
            if draw:
                # Draw the bounding box around the hand on the output image.
                cv2.rectangle(output_image, (x1, y1), (x2, y2), (155, 0, 255), 5, cv2.LINE_8)

                # Write the classified label of the hand below the bounding box drawn.
                cv2.putText(output_image, label, (x1 + 15, y2 + 58), cv2.FONT_HERSHEY_COMPLEX, 1.8, (255, 255, 0), 2,
                            cv2.LINE_AA)

        # Check if the output image is specified to be displayed.
        if display:

            # Display the output image.
            plt.figure(figsize=[10, 10])
            plt.imshow(output_image[:, :, ::-1]);
            plt.title("Output Image");
            plt.axis('off');

        # Otherwise
        else:

            # Return the output image and the landmarks dictionary.
            return output_image, output_landmarks

    def customLandmarksAnnotation(image, landmark_dict):
        '''
        This function draws customized landmarks annotation utilizing the z-coordinate (depth) values of the hands.
        Args:
            image:         The image of the hands on which customized landmarks annotation of the hands needs to be drawn.
            landmark_dict: The dictionary that stores the hand(s) landmarks as different elements with keys as hand
                           types(i.e., left and right).
        Returns:
            output_image: The image of the hands with the customized annotation drawn.
            depth:        A dictionary that contains the average depth of all landmarks of the hand(s) in the image.
        '''

        # Create a copy of the input image to draw annotation on.
        output_image = image.copy()

        # Initialize a dictionary to store the average depth of all landmarks of hand(s).
        depth = {}

        # Initialize a list with the arrays of indexes of the landmarks that will make the required
        # line segments to draw on the hand.
        segments = [np.arange(0, 5), np.arange(5, 9), np.arange(9, 13), np.arange(13, 17), np.arange(17, 21),
                    np.arange(5, 18, 4), np.array([0, 5]), np.array([0, 17])]

        # Iterate over the landmarks dictionary.
        for hand_type, hand_landmarks in landmark_dict.items():

            # Get all the z-coordinates (depth) of the landmarks of the hand.
            depth_values = np.array(hand_landmarks)[:, -1]

            # Calculate the average depth of the hand.
            average_depth = int(sum(depth_values) / len(depth_values))

            # Get all the x-coordinates of the landmarks of the hand.
            x_values = np.array(hand_landmarks)[:, 0]

            # Get all the y-coordinates of the landmarks of the hand.
            y_values = np.array(hand_landmarks)[:, 1]

            # Initialize a list to store the arrays of x and y coordinates of the line segments for the hand.
            line_segments = []

            # Iterate over the arrays of indexes of the landmarks that will make the required line segments.
            for segment_indexes in segments:
                # Get an array of a line segment coordinates of the hand.
                line_segment = np.array([[int(x_values[index]), int(y_values[index])] for index in segment_indexes])

                # Append the line segment coordinates into the list.
                line_segments.append(line_segment)

            # Check if the average depth of the hand is less than 0.
            if average_depth < 0:

                # Set the thickness of the line segments of the hand accordingly to the average depth.
                line_thickness = int(np.ceil(0.1 * abs(average_depth))) + 2

                # Set the thickness of the circles of the hand landmarks accordingly to the average depth.
                circle_thickness = int(np.ceil(0.1 * abs(average_depth))) + 3

            # Otherwise.
            else:

                # Set the thickness of the line segments of the hand to 2 (i.e. the minimum thickness we are specifying).
                line_thickness = 2

                # Set the thickness of the circles to 3 (i.e. the minimum thickness)
                circle_thickness = 3

            # Draw the line segments on the hand.
            cv2.polylines(output_image, line_segments, False, (100, 250, 55), line_thickness)

            # Write the average depth of the hand on the output image.
            cv2.putText(output_image, 'Depth (in Z direction): {}'.format(average_depth), (280, 120 + 100 + 300),
                        cv2.FONT_HERSHEY_COMPLEX, 2.8, (255, 25, 0), 3, cv2.LINE_AA)

            # Iterate over the x and y coordinates of the hand landmarks.
            for x, y in zip(x_values, y_values):
                # Draw a circle on the x and y coordinate of the hand.
                cv2.circle(output_image, (int(x), int(y)), circle_thickness, (55, 55, 250), -1)

            # Store the calculated average depth in the dictionary.
            depth[hand_type] = average_depth

        # Return the output image and the average depth dictionary of the hand(s).
        return output_image, depth

#




def main():
        pTime = 0
        cTime = 0
        cap = cv2.VideoCapture(2)
        detector = handDetector()

        while True:
            success, img = cap.read()
            img = detector.findHands(img, draw=True)
            # lmList = detector.findPosition(img, draw=False)
            lmList,bbox = detector.findPosition(img, draw=True)
            if len(lmList) != 0:
                pass
                # print(lmList[8])

            cTime = time.time()
            fps = 1 / (cTime - pTime)
            pTime = cTime

            cv2.putText(img, str(int(fps)), (10, 70), cv2.FONT_HERSHEY_PLAIN, 5,(255, 0, 255), 3)
            cv2.imshow("Image", img)
            key = cv2.waitKey(1)
            if key == 27 or key == ord('q'):
                break
if __name__ == '__main__':
    main()